#include "group.h"


group::group(int numDimensions){
	this->numDimensions = numDimensions;
	extentCalculated = false;
	numMicroPointsCalculated = false;
}

group::group(){
}

group::group(const group &gp){
	numDimensions = gp.numDimensions;
	extentCalculated = gp.extentCalculated;
	numMicroPointsCalculated = gp.numMicroPointsCalculated;
	if (extentCalculated){
		extent = new float* [numDimensions];
		for (int d=0; d<numDimensions; d++){
			extent[d] = new float[2];
			extent[d][0] = gp.extent[d][0];
			extent[d][1] = gp.extent[d][1]; 
		}
	}
	if (numMicroPointsCalculated){
		numMicroPoints = gp.numMicroPoints;
	}
	for (uint i=0; i<gp.points.size(); i++){
		points.push_back(gp.points[i]);
	}
	
//	vector<point>::iterator iter; // when points is vector
////	set<point>::iterator iter; // when points is set
//	for (iter=gp.points.begin(); iter!=gp.points.end(); iter++){
//		points.push_back(*iter); // when points is vector
////		points.insert(*iter); // when points is set
//	}

}

group& group::operator= (const group& gp){
	this->numDimensions = gp.numDimensions;
	this->extentCalculated = gp.extentCalculated;
	this->numMicroPointsCalculated = gp.numMicroPointsCalculated;
	if (this->extentCalculated){
		this->extent = new float* [numDimensions];
		for (int d=0; d<numDimensions; d++){
			this->extent[d] = new float[2];
			this->extent[d][0] = gp.extent[d][0];
			this->extent[d][1] = gp.extent[d][1]; 
		}
	}
	if (this->numMicroPointsCalculated){
		this->numMicroPoints = gp.numMicroPoints;
	}
	for (uint i=0; i<gp.points.size(); i++){
		this->points.push_back(gp.points[i]);
	}
//	vector<point>::iterator iter; // when points is vector
////	set<point>::iterator iter; // when points is set
//	for (iter=gp.points.begin(); iter!=gp.points.end(); iter++){
//		this->points.push_back(*iter); // when points is vector
////		this->points.insert(*iter); // when points is set
//	}
	
	return *this;
}


group::~group(){
	if (extentCalculated){
		for (int d=0; d<numDimensions; d++){
			delete[] extent[d];
		}
		delete[] extent;
	}
}

bool group::operator==(const group& gp) const{
	return false;
}

void group::insertPoint(point& po){
	if (extentCalculated){
		for (int d=0; d<numDimensions; d++){
			delete[] extent[d];
		}
		delete[] extent;
	}	
	extentCalculated = false;
	numMicroPointsCalculated = false;
	points.push_back(po); // when points is vector
//	points.insert(po); // when points is set
}

float** group::getExtent(){
	if (!extentCalculated){ // calculate extent
		extentCalculated = true;
		extent = new float* [numDimensions];
		for (int d=0; d<numDimensions; d++){
			extent[d] = new float[2];
		}
		vector<point>::iterator iter; // when points is vector
//		set<point>::iterator iter; // when points is set
		
		for(iter=points.begin(); iter!=points.end(); iter++){
			for (int d=0; d<numDimensions; d++){
				if (iter==points.begin()){
					extent[d][0] = iter->p[d];
					extent[d][1] = iter->p[d];
				}
				else{
					if (iter->p[d]<extent[d][0]){
						extent[d][0] = iter->p[d];
					}
					if (iter->p[d]>extent[d][1]){
						extent[d][1] = iter->p[d];
					}
				}
			}
		}
	}
	return extent;
}

int group::getNumMicroPoints(){
	if (!numMicroPointsCalculated){ // count points in MicroTable
		numMicroPointsCalculated = true;
		numMicroPoints = 0;
		vector<point>::iterator iter; // when points is vector
//		set<point>::iterator iter; // when points is set
		for(iter=points.begin(); iter!=points.end(); iter++){
			if (iter->inMicro) numMicroPoints++;
		}
	}
	return numMicroPoints;
}

float group::getPerimeter(){
	float perimeter = 0;
	float** extent = getExtent();
	for (int d=0; d<numDimensions; d++){
		perimeter += extent[d][1] - extent[d][0]; 
	}
	return perimeter;
}

float group::getNormalPerimeter(float* sizeWorld){
	float perimeter = 0;
	float** extent = getExtent();
	for (int d=0; d<numDimensions; d++){
		perimeter += (extent[d][1] - extent[d][0])/sizeWorld[d]; 
	}
	return perimeter;
}

ostream& operator<<(ostream& os, const groupPtr &gp){
	float** extent = gp->getExtent();
	os << "numPoints=" << gp->points.size() << " numMicroPoint=" << gp->getNumMicroPoints();
	os << " (unnormalized) perimeter=" << gp->getPerimeter() << endl;
	os << "extent = |";
	for (int d=0; d<gp->numDimensions; d++){
		os << extent[d][0] << " -- " << extent[d][1] << "| ";
	}
	os << endl;
	vector<point>::iterator iter; // when points is vector
//	set<point>::iterator iter;
	for (iter=gp->points.begin(); iter!=gp->points.end(); iter++){
		os << *iter << " ";
	}
	os << endl;
	return os;
}

void calculateMetrics(vector<group> groups, float* sizeWorld){	
	float normalPerimeter = 0;
	float sizeTimesPerimeter = 0;
	float discernability = 0;
	
	vector<group>::iterator iter;
	
	int totalMicroPoints = 0;
	
	for (iter=groups.begin(); iter!=groups.end(); iter++){
		int numMicroPoints = iter->getNumMicroPoints();
		totalMicroPoints += numMicroPoints;
		discernability += numMicroPoints*numMicroPoints;
		float perimeter = iter->getNormalPerimeter(sizeWorld);
//		if (perimeter==0) cerr << "hurray! group with 0 perimeter" << endl;
		normalPerimeter += perimeter;
		sizeTimesPerimeter += perimeter*numMicroPoints;
	}
	
	
	cout << "calculateMetrics:::: " << groups.size() << " groups" << endl;
	cout << "totalMicroPoints = " << totalMicroPoints << endl; 
	cout << "discernability = " << discernability << endl;
	cout << "total perimeter = " << normalPerimeter << endl;
	cout << "average perimeter = " << normalPerimeter/groups.size() << endl;
	cout << "total (perimeter)*(numMicroPoint) = " << sizeTimesPerimeter << endl;
	
}


