#include "hilKAnom.h"



HilKAnom::HilKAnom(char* inputfile){
	groupsCalculated = false;
	cout << "HilKAnom:: reading from " << inputfile << endl;
	fstream fin(inputfile, ios::in);
	
	fin >> numDimensions; // first read dimensionality
	float p[numDimensions];
	float* mini = new float[numDimensions];
	float* maxi = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		mini[i] = FLT_MAX;
		maxi[i] = FLT_MIN;
	}
	
	while(!fin.eof()){
		for (int i=0; i<numDimensions; i++){
			fin >> p[i];
			if (p[i]<mini[i]){
				mini[i] = p[i];
			}
			if (p[i]>maxi[i]){
				maxi[i] = p[i];
			}
		}
		point *po;
		po = new point(numDimensions, p, true);
		points.push_back(*po);
	}
	fin.close();
	
	sizeWorld = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = maxi[i] - mini[i];
	}
	delete[] mini;
	delete[] maxi;
	
	cerr << "...done reading " << points.size() << " points" << endl;
	
	kdtree = new KDTree(inputfile, 1); // build a KDTree...
//	cout << "KDTree built" << endl;
	
}


HilKAnom::~HilKAnom(){
	delete[] sizeWorld;
	delete kdtree;
}


float* HilKAnom::getSizeWorld(){
	return sizeWorld;
}


vector<group> HilKAnom::getGroups(int K){
	if (!groupsCalculated || this->K!=K){
		groups.clear();
		
		groupsCalculated = true;
		this->K = K;
		
		vector<point>::iterator iter;
		iter = points.begin();
		
		bool groupOK = false;
		while(iter!=points.end()){
			int count = 0;
			group* gp = new group(numDimensions);
			
			groupOK = false;
			while(!groupOK && iter!=points.end()){
				gp->insertPoint(*iter);
				count++;
				iter++;
				if (count>=K){
					groupOK = true;
				}
			}
			
			if (groupOK){
				float** extent = gp->getExtent();
				vector<pointPtr> result = kdtree->rangeQuery(extent);
				for (uint j=0; j<result.size(); j++){
					gp->insertPoint(*(result[j]));
				}
				groups.push_back(*gp);
			}
			else{ // merge this with last group
				group lastGroup = groups[groups.size()-1];
				groups.pop_back();
				vector<point>::iterator it;
				for (it=gp->points.begin(); it!=gp->points.end(); it++){
					point po = *it;
					lastGroup.insertPoint(po);
				}
				float** extent = lastGroup.getExtent();
				vector<pointPtr> result = kdtree->rangeQuery(extent);
				for (uint j=0; j<result.size(); j++){
					lastGroup.insertPoint(*(result[j]));
				}
				groups.push_back(lastGroup);
			}
			
			
		}
	}
	return groups;
}


