#include "kdtree.h"


using namespace std;

KDTree::KDTree(char *inputfile, int K){
	groupsCalculated = false;
	float *p;
	vector<float *> v;
	cerr << "KDTree:: reading from " << inputfile << endl;
	fstream fin(inputfile, ios::in);

	fin >> numDimensions; // first read dimensionality
	while(!fin.eof()){
		p = new float[numDimensions];
		for (int i=0; i<numDimensions; i++){
			fin >> p[i];
		}
		v.push_back(p);
	}
	fin.close();
	cerr << "... done reading " << v.size() << " points" << endl;

	this->K = K;

//	cout << "K=" << this->K << endl;

	int numPoints = v.size();
	pointPtr *points = new pointPtr[numPoints];
	for (int i=0; i<numPoints; i++){
		points[i] = new point(numDimensions, v[i], true);
	}
	v.clear();

	root = new KDNode(numDimensions, numPoints, points, 0);
	root->id = 1;
	root->calculateExtentMBR(1);

	nodes[1] = root;

	calculateSizeWorld();

	cerr << "KDTree:: gonna do the splitting..." << endl;
	recursiveSplit(root);
	cerr << "...done with the splitting" << endl;
}


KDTree::KDTree(vector<pointPtr> poiset, int K){
	groupsCalculated = false;
	numDimensions = poiset[0]->numDimensions;
	this->K = K;

	int numPoints = poiset.size();
	pointPtr *points = new pointPtr[numPoints];
	copy(poiset.begin(), poiset.end(), points);

	root = new KDNode(numDimensions, numPoints, points, 0);
	root->id = 1;
	root->calculateExtentMBR(1);

	nodes[1] = root;

	calculateSizeWorld();

	cerr << "KDTree:: gonna do the splitting..." << endl;
	recursiveSplit(root);
	cerr << "...done with the splitting" << endl;
}


void KDTree::recursiveSplit(KDNodePtr node){
//	cerr << "recursiveSplit @ " << node->id << endl;
	if (node->numPoints>=2*K){
		KDNodePtr left, right;
		bool hasSplit = node->splitNode(left, right);
//		cout << "splitDimension = " << node->splitDimension << " hasSplit = " << hasSplit << endl;

		if (hasSplit){
			nodes[left->id] = left;
			nodes[right->id] = right;
			left->calculateNumMicroPoints();
			right->calculateNumMicroPoints();



			if (left->numMicroPoints==0){
//				cerr << "no microPoints in left" << endl;
				nodes.erase(left->id);
				delete left;
				node->left = NULL;
			}
			else{
				recursiveSplit(left);
			}

			if (right->numMicroPoints==0){
//				cerr << "no microPoints in right" << endl;
				nodes.erase(right->id);
				delete right;
				node->right = NULL;
			}
			else{
				recursiveSplit(right);
			}



		}
	}
}


void KDTree::calculateSizeWorld(){
	sizeWorld = new float[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = nodes[1]->extent[i][1] - nodes[1]->extent[i][0];
	}
}

//float* KDTree::getSizeWorld(){
//	float* sizeWorld = new float[numDimensions];
//	for (int i=0; i<numDimensions; i++){
//		sizeWorld[i] = nodes[1]->extent[i][1] - nodes[1]->extent[i][0];
//	}
//	return sizeWorld;
//}


vector<group> KDTree::getGroups(){
	if (!groupsCalculated){
		groups.clear();
		groupsCalculated = true;

		map<int, KDNodePtr>::iterator iter;
		for (iter=nodes.begin(); iter!=nodes.end(); iter++){
			if (!iter->second->leafNode){
				continue; // this is not a leaf node
			}
			group* gp = new group(numDimensions);

			for (int i=0; i<iter->second->numPoints; i++){
				gp->insertPoint(*(iter->second->points[i]));
			}
			groups.push_back(*gp);
		}
	}
	return groups;
}

void KDTree::writeGroups(char *inputfile){
	ostringstream kdfied;
	kdfied << "K" << K << "_" << inputfile;
	string writefile = kdfied.str();
	ofstream fout(writefile.c_str(), ios::out);

	map<int, KDNodePtr>::iterator iter;
	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->leafNode){
			continue; // this is not a leaf node
		}
		fout << iter->second << endl;
	}

	fout.close();
}


void KDTree::printMetrics(){
	map<int, KDNodePtr>::iterator iter;

	float normalPerimeter = 0;
	float sizeTimesPerimeter = 0;
	float discernability = 0;
	int numGroups = 0;

	float sizeWorld[numDimensions];
	for (int i=0; i<numDimensions; i++){
		sizeWorld[i] = nodes[1]->extent[i][1] - nodes[1]->extent[i][0];
	}

	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->leafNode){
			continue; // this is not a leaf node
		}
		numGroups++;
		discernability += iter->second->numMicroPoints*iter->second->numMicroPoints;

		float thisPerimeter = 0;
		for (int i=0; i<numDimensions; i++){
//			cout << "... " << (iter->second->extent[i][1] - iter->second->extent[i][0])/sizeWorld[i] << endl;
			float diff = iter->second->extent[i][1] - iter->second->extent[i][0];
			thisPerimeter += diff/sizeWorld[i];
		}
		normalPerimeter += thisPerimeter;
		sizeTimesPerimeter += thisPerimeter*iter->second->numMicroPoints;
	}

	cout << "there are " << numGroups << " groups" << endl;
	cout << "discernability = " << discernability << endl;
//	cout << "normalized average equivalence class size = " << ((float) K*nodes[1]->numMicroPoints)/numGroups << endl;
	cout << "total perimeter = " << normalPerimeter << endl;
	cout << "total (perimeter)*(numMicroPoint) = " << sizeTimesPerimeter << endl;

}


/*
void KDTree::boostWithPublicData(KDNodePtr node, MinPerTreePtr pub){
//	cout << "boostWithPublicData" << endl;
	if (!node->leafNode){
		boostWithPublicData(node->left, pub);
		boostWithPublicData(node->right, pub);
	}
	else{ // do the range query
//		cout << "do the range query" << endl;
		vector<pointPtr> result = pub->rangeQuery(node->mbr);
		vector<pointPtr> newPoints = result;
//		cout << "query ok" << endl;
		vector<pointPtr>::iterator resIter;
		for (resIter=newPoints.begin(); resIter!=newPoints.end(); resIter++){
//			cout << "."; fflush(stdout);
			(*resIter)->inMicro = false;
		}
//		cout << "ok" << endl;
//		cout << "gonna create thesePoints with " << node->numPoints << " points" << endl;
//		vector<pointPtr> thesePoints;
//		cout << "before inserting thesePoints.size()=" << thesePoints.size() << endl;
//		for (int i=0; i<node->numPoints; i++){
//			cout << *(node->points[i]) << " ";
//			thesePoints.push_back(node->points[i]);
//		}
//		cout << endl;
		vector<pointPtr> thesePoints(node->numPoints);
		copy(node->points, node->points+node->numPoints, thesePoints.begin());

//		cout << "newPoints.size()=" << newPoints.size() << endl;
//		cout << "thesePoints.size()=" << thesePoints.size() << endl;

		vector<pointPtr> joinedPoints = SortMerge(newPoints, thesePoints);

//		cout << "joinedPoints.size()=" << joinedPoints.size() << endl;

		delete[] node->points;

		node->numPoints = joinedPoints.size();
		node->points = new pointPtr[node->numPoints];
		copy(joinedPoints.begin(), joinedPoints.end(), node->points);
		node->calculateNumMicroPoints();
//		cout << "numMicroPoints=" << node->numMicroPoints << endl;
		recursiveSplit(node);

	}
}
*/

void KDTree::boostWithPublicData(KDNodePtr node, KDTreePtr pub){
//	cout << "boostWithPublicData" << endl;
	if (!node->leafNode){
		boostWithPublicData(node->left, pub);
		boostWithPublicData(node->right, pub);
	}
	else{ // do the range query
		node->boostedLeafNode = true;
//		cout << "do the range query" << endl;
		vector<pointPtr> result = pub->rangeQuery(node->mbr);
//		cerr << "before numPoints=" << node->numPoints << " numMicroPoints=" << node->numMicroPoints << endl;
//		cerr << result.size() << " points returned" << endl;
		vector<pointPtr> newPoints = result;
//		cout << "query ok" << endl;
		vector<pointPtr>::iterator resIter;
		for (resIter=newPoints.begin(); resIter!=newPoints.end(); resIter++){
//			cout << "."; fflush(stdout);
			(*resIter)->inMicro = false;
		}
//		cout << "ok" << endl;
//		cout << "gonna create thesePoints with " << node->numPoints << " points" << endl;
//		vector<pointPtr> thesePoints;
//		cout << "before inserting thesePoints.size()=" << thesePoints.size() << endl;
//		for (int i=0; i<node->numPoints; i++){
//			cout << *(node->points[i]) << " ";
//			thesePoints.push_back(node->points[i]);
//		}
//		cout << endl;
		vector<pointPtr> thesePoints(node->numPoints);
		copy(node->points, node->points+node->numPoints, thesePoints.begin());

//		cout << "newPoints.size()=" << newPoints.size() << endl;
//		cout << "thesePoints.size()=" << thesePoints.size() << endl;

		vector<pointPtr> joinedPoints = SortMerge(newPoints, thesePoints);

//		cout << "joinedPoints.size()=" << joinedPoints.size() << endl;

		delete[] node->points;

		node->numPoints = joinedPoints.size();
		node->points = new pointPtr[node->numPoints];
		copy(joinedPoints.begin(), joinedPoints.end(), node->points);
		node->calculateNumMicroPoints();
//		cerr << "after numPoints=" << node->numPoints << " numMicroPoints=" << node->numMicroPoints << endl;
		recursiveSplit(node);

	}
}


void KDTree::boostWithPublicData(KDTreePtr pub){
	groupsCalculated = false;
	map<int, KDNodePtr>::iterator iter;
	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (iter->second->leafNode){
			iter->second->boostedLeafNode = true;
		}
	}

//	int lStart = -1;
//	int lEnd = -1;
////	map<int, KDNodePtr>::iterator iter;
//	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
//		if (!iter->second->leafNode){
//			continue; // this is not a leaf node
//		}
//		if (lStart==-1) lStart = iter->first;
//		lEnd = iter->first;
//	}


	for (iter=nodes.begin(); iter!=nodes.end(); iter++){
		if (!iter->second->boostedLeafNode){
			continue; // this is not a leaf node that is going to be boosted
		}

		vector<pointPtr> result = pub->rangeQuery(iter->second->mbr);
		vector<pointPtr>::iterator resIter;
		for (resIter=result.begin(); resIter!=result.end(); resIter++){
			(*resIter)->inMicro = false;
		}
//		cout << "range:: | ";
//		for (int j=0; j<numDimensions; j++){
//			cout << iter->second->extent[j][0] << " -- " << iter->second->extent[j][1] << " | ";
//		}
//		cout << endl;
//		cout << "for node " << iter->second->id << " rangeQueryStr returned " << result.size() << " results" << endl;
//		for (uint i=0; i<result.size(); i++){
//			cout << *(result[i]) << " ";
//		}
//		cout << endl;

		for (int i=0; i<iter->second->numPoints; i++){
			pointPtr mtp = iter->second->points[i];
			for (resIter=result.begin(); resIter!=result.end(); resIter++){
				if ( *mtp == *(*resIter) ){
					(*resIter)->inMicro = true;
					break;
				}
			}
		}

		pointPtr *newpoints = new pointPtr[result.size()];
		for (uint i=0; i<result.size(); i++){
			newpoints[i] = new point(*(result[i]));
		}

		for (int i=0; i<iter->second->numPoints; i++){
			delete iter->second->points[i];
		}
		delete[] iter->second->points;

		iter->second->points = newpoints;
		iter->second->numMicroPoints = iter->second->numPoints;
		iter->second->numPoints = result.size();

		recursiveSplit(iter->second);

	}

}


vector<pointPtr> KDTree::rangeQuery(float **query){
	vector<pointPtr> result;
//	cout << "got range query " << getString(query, numDimensions) << endl;
	recRangeQuery(root, query, result);
//	cout << "range query returned" << endl;

	return result;
}

void KDTree::recRangeQuery(KDNodePtr node, float **query, vector<pointPtr>& res){
//	cout << "visiting node " << node->id << " :: " << endl;
//	cout << "visiting " << node << endl;
//	cout << node << endl;
	if (node->leafNode){
//	if ( node->left==NULL && node->right==NULL ){ // leaf node
//		cout << "leaf node numPoints=" << node->numPoints << endl;
		for (int i=0; i<node->numPoints; i++){
//			cout << i << " "; fflush(stdout);
			if ( inside(node->points[i]->p, query, numDimensions) ){ // point inside
//				cout << "in "; fflush(stdout);
				res.push_back(node->points[i]);
//				cout << "in" << *(node->points[i]) << " ";
			}
			else{
//				cout << "out " << getString(node->points[i], numDimensions) << endl;
			}
		}
	}
//	cout << endl;
//	cout << "gonna try to descend" << endl;
	if ( node->left!=NULL && overlaps(query, node->left->mbr, numDimensions) ){
		recRangeQuery(node->left, query, res); // descend
	}
	if ( node->right!=NULL && overlaps(query, node->right->mbr, numDimensions) ){
		recRangeQuery(node->right, query, res); // descend
	}
//	cout << "descend returned" << endl;

}

vector<pointPtr> KDTree::rangeQueryStr(float **query){
	vector<pointPtr> result;

	recRangeQueryStr(root, query, result);

	return result;
}

void KDTree::recRangeQueryStr(KDNodePtr node, float **query, vector<pointPtr>& res){
//	cout << "visiting node " << node->id << endl;
	if ( node->left==NULL && node->right==NULL ){ // leaf node
//		cout << "leaf node numPoints=" << node->numPoints << endl;
		for (int i=0; i<node->numPoints; i++){
			if ( insideStr(node->points[i]->p, query, numDimensions) ){ // point inside
				res.push_back(node->points[i]);
//				cout << "in " << getString(node->points[i], numDimensions) << endl;
			}
			else{
//				cout << "out " << getString(node->points[i], numDimensions) << endl;
			}
		}
	}
	if ( node->left!=NULL && overlapsStr(query, node->left->extent, numDimensions) ){
		recRangeQueryStr(node->left, query, res); // descend
	}
	if ( node->right!=NULL && overlapsStr(query, node->right->extent, numDimensions) ){
		recRangeQueryStr(node->right, query, res); // descend
	}

}


void KDTree::countQuery(float **query, double& estimate, double& actual){
	recCountQuery(root, query, estimate, actual);
}

void KDTree::recCountQuery(KDNodePtr node, float **query, double& estimate, double& actual){
//	cout << "recCountQuery @ " << node->id << endl;

	if (node->leafNode){
		float **inter;
		inter = intersect(query, node->mbr, numDimensions);
		double interVol = 1, nodeVol = 1;
		for (int i=0; i<numDimensions; i++){
			interVol *= inter[i][1] - inter[i][0];
			nodeVol *= node->mbr[i][1] - node->mbr[i][0];
			delete inter[i];
		}
		delete inter;
		if (interVol==0) interVol = 1;
		if (nodeVol==0) nodeVol = 1;
//		cout << "interVol=" << interVol << " nodeVol=" << nodeVol << endl;
//		cout << "estimated MT tuples " << (interVol/nodeVol) * node->numMicroPoints << endl;
		estimate += (interVol/nodeVol) * node->numMicroPoints;


		double numMicro = 0;
		for (int i=0; i<node->numPoints; i++){
			if (node->points[i]->inMicro &&  insideStr(node->points[i]->p, query, numDimensions) ){ // point inside
				numMicro++;
			}
		}
//		cout << "actual MT tuples " << numMicro << endl;


		actual += numMicro;

	}
	else {
		if (node->left != NULL  &&  overlaps(query, node->left->mbr, numDimensions)){
			recCountQuery(node->left, query, estimate, actual);
		}
		if (node->right != NULL  &&  overlaps(query, node->right->mbr, numDimensions)){
			recCountQuery(node->right, query, estimate, actual);
		}
	}

}




void KDTree::countQueryREF(float **query, double& estimate, double& actual){
	recCountQueryREF(root, query, estimate, actual);
}


void KDTree::recCountQueryREF(KDNodePtr node, float **query, double& estimate, double& actual){
//	cout << "recCountQueryREF @ " << node->id << endl;

	if (node->leafNode){
		float **inter;
		inter = intersect(query, node->mbr, numDimensions);
		double interVol = 1, nodeVol = 1;
		for (int i=0; i<numDimensions; i++){
			interVol *= inter[i][1] - inter[i][0];
			nodeVol *= node->mbr[i][1] - node->mbr[i][0];
			delete inter[i];
		}
		delete inter;
		if (interVol==0) interVol = 1;
		if (nodeVol==0) nodeVol = 1;
//		cout << "interVol=" << interVol << " nodeVol=" << nodeVol << endl;
//		cout << "estimated MT tuples " << (interVol/nodeVol) * node->perceivedMicroTuples << endl;
		estimate += (interVol/nodeVol) * node->perceivedMicroTuples;


		double numMicro = 0;
		for (int i=0; i<node->numPoints; i++){
			if (node->points[i]->inMicro &&  insideStr(node->points[i]->p, query, numDimensions) ){ // point inside
				numMicro++;
			}
		}
//		cout << "actual MT tuples " << numMicro << endl;


		actual += numMicro;

	}
	else {
		if (node->left != NULL  &&  overlaps(query, node->left->mbr, numDimensions)){
			recCountQueryREF(node->left, query, estimate, actual);
		}
		if (node->right != NULL  &&  overlaps(query, node->right->mbr, numDimensions)){
			recCountQueryREF(node->right, query, estimate, actual);
		}
	}

}



void KDTree::calculatePerceivedMicroTuples(){
	recCalculatePerceivedMicroTuples(root, 1);
}

void KDTree::recCalculatePerceivedMicroTuples(KDNodePtr node, double ratio){
//	cout << "recCalculatePerceivedMicroTuples @ " << node->id << endl;

//	if (node->parent != NULL && node->numMicroPoints < K && node->parent->numMicroPoints >= K){
//		ratio = ((double) node->parent->numMicroPoints) / node->parent->leafTuples;
//		cout << "parent node with id " << node->parent->id << " was boosted " << node->parent->boostedLeafNode << endl;
//	}

	if (node->boostedLeafNode){
		ratio = ((double) node->numMicroPoints) / node->leafTuples;
	}


//	cout << "ratio is " << ratio << endl;

	if (node->leafNode){
		node->perceivedMicroTuples = ratio * node->leafTuples;
//		cout << "leaf node " << node->id << " has perceivedMicroTuples=" << node->perceivedMicroTuples << " leafTuples=" << node->leafTuples << " and numMicroPoints=" << node->numMicroPoints<< endl;
	}
	else {
		if (node->left != NULL){
			recCalculatePerceivedMicroTuples(node->left, ratio);
		}
		if (node->right != NULL){
			recCalculatePerceivedMicroTuples(node->right, ratio);
		}
	}
}


void KDTree::countPerceivedMicroTuples(){
	double count = 0;
	recCountPerceivedMicroTuples(root, count);
	cerr << "countPerceivedMicroTuples is " << count << endl;
}

void KDTree::recCountPerceivedMicroTuples(KDNodePtr node, double& count){
//	cout << "recCountPerceivedMicroTuples @ " << node->id << endl;

	if (node->leafNode){
		count += node->perceivedMicroTuples;
//		cout << " + " << node->perceivedMicroTuples << " tuples" << endl;
	}
	else {
		if (node->left != NULL){
			recCountPerceivedMicroTuples(node->left, count);
		}
		if (node->right != NULL){
			recCountPerceivedMicroTuples(node->right, count);
		}
	}
}





void KDTree::calculateLeafTuples(){
	recCalculateLeafTuples(root);
}


void KDTree::recCalculateLeafTuples(KDNodePtr node){
//	cout << "recCalculateLeafTuples @ " << node->id << endl;
	if (node->leafNode){

		if (node->numMicroPoints >= K){
			node->leafTuples = node->numMicroPoints;
		}
		else {
			node->leafTuples = K;
		}
//		cout << "leaf node tuples " << node->leafTuples << endl;
	}
	else {
		node->leafTuples = 0;
		if (node->left != NULL){
			recCalculateLeafTuples(node->left);
			node->leafTuples += node->left->leafTuples;
		}
		if (node->right != NULL){
			recCalculateLeafTuples(node->right);
			node->leafTuples += node->right->leafTuples;
		}
//		cout << "--- node " << node->id << " has tuples " << node->leafTuples << endl;
	}
}



ostream& operator<<(ostream& os, const KDTreePtr& t){
	map<int, KDNodePtr>::iterator iter;

	cout << endl;
	for(iter = t->nodes.begin(); iter!=t->nodes.end(); iter++){
        cout << "map id=" << iter->first << " " << iter->second << endl;
    }

	return os;
}

