#include "minpernode.h"


MinPerNode::MinPerNode(int numDimensions, int numPoints, pointPtr* points, int splitDimension){
	this->numDimensions = numDimensions;
	this->numPoints = numPoints;
	this->points = points;
	this->splitDimension = splitDimension;
	this->numMicroPoints = calculateNumMicroPoints();
	splitValue = -2007;
	left = NULL;
	right = NULL;
	parent = NULL;
	leafNode = true;
	boostedLeafNode = false;
}

MinPerNode::~MinPerNode(){
	for (int i=0; i<numDimensions; i++){
		delete[] extent[i];
		delete[] mbr[i];
	}
	delete[] extent;
	delete[] mbr;
	for (int i=0; i<numPoints; i++){
		delete points[i];
	}
	delete[] points;
}

int MinPerNode::calculateNumMicroPoints(){
	numMicroPoints = 0;
	for (int i=0; i<numPoints; i++){
		if (points[i]->inMicro){
			numMicroPoints++;
		}
	}
	return numMicroPoints;
}

void MinPerNode::calculateExtentMBR(float pad){
	mbr = new float * [numDimensions];
	extent = new float * [numDimensions];
	vector<float> temp;
	for (int d=0; d<numDimensions; d++){
		mbr[d] = new float[2];
		extent[d] = new float[2];
		for (int i=0; i<numPoints; i++){
			temp.push_back(points[i]->p[d]);
		}
		
		vector<float>::iterator it = min_element(temp.begin(), temp.end());
		
		mbr[d][0] = *min_element(temp.begin(), temp.end()) - pad;
		mbr[d][1] = *max_element(temp.begin(), temp.end()) + pad;
		
		if (parent==NULL){ // extent = mbr
			extent[d][0] = mbr[d][0];
			extent[d][1] = mbr[d][1];
		}
		else{
			int parentSplitDimension = (splitDimension+numDimensions-1)%numDimensions;
			float parentSplitValue = parent->splitValue;
//			cout << "parentSplitDimension=" << parentSplitDimension << " parentSplitValue=" << parentSplitValue << endl;
			if (d!=parentSplitDimension){
				extent[d][0] = parent->extent[d][0];
				extent[d][1] = parent->extent[d][1];
			}
			else{
				// find out if I'm left or right child
				if (points[0]->p[d]<=parentSplitValue){
//					cout << "I'm a left child" << endl;
					extent[d][0] = parent->extent[d][0];
					extent[d][1] = parentSplitValue;
				}
				else{
//					cout << "I'm a right child" << endl;
					extent[d][0] = parentSplitValue;
					extent[d][1] = parent->extent[d][1];
				}
			}
		}
		
		temp.clear();
	}
}


void MinPerNode::calculateExtentMBR(){
	mbr = new float * [numDimensions];
	extent = new float * [numDimensions];
	vector<float> temp;
	for (int d=0; d<numDimensions; d++){
		mbr[d] = new float[2];
		extent[d] = new float[2];
		for (int i=0; i<numPoints; i++){
			temp.push_back(points[i]->p[d]);
		}
		
		vector<float>::iterator it = min_element(temp.begin(), temp.end());
		
		mbr[d][0] = *min_element(temp.begin(), temp.end());
		mbr[d][1] = *max_element(temp.begin(), temp.end());
		
		if (parent==NULL){ // extent = mbr
			extent[d][0] = mbr[d][0];
			extent[d][1] = mbr[d][1];
		}
		else{
			int parentSplitDimension = (splitDimension+numDimensions-1)%numDimensions;
			float parentSplitValue = parent->splitValue;
//			cout << "parentSplitDimension=" << parentSplitDimension << " parentSplitValue=" << parentSplitValue << endl;
			if (d!=parentSplitDimension){
				extent[d][0] = parent->extent[d][0];
				extent[d][1] = parent->extent[d][1];
			}
			else{
				// find out if I'm left or right child
				if (points[0]->p[d]<=parentSplitValue){
//					cout << "I'm a left child" << endl;
					extent[d][0] = parent->extent[d][0];
					extent[d][1] = parentSplitValue;
				}
				else{
//					cout << "I'm a right child" << endl;
					extent[d][0] = parentSplitValue;
					extent[d][1] = parent->extent[d][1];
				}
			}
		}
		
		temp.clear();
	}
}


bool MinPerNode::splitNodeNEW(MinPerNodePtr &left, MinPerNodePtr &right, float* theSizeWorld, int K){
//	cerr << "splitNodeNEW (numPoints=" << numPoints << ") "; fflush(stderr);
//	if (numPoints>10*K) return fastSplitNode(left, right); // OPTIMIZATION
	
//	cout << "called splitNode for id=" << this->id << endl;
	float* sizeWorld = theSizeWorld;
	
	
	int bestDimToSplit = -1;
	float bestDimSplitValue;
	float bestScore = FLT_MAX;
	int bestLeftNumPoints;
	
	for (int thisD=0; thisD<numDimensions; thisD++){
//		cout << "checking dimension " << thisD << endl;
		
		
		orderByDim *myorder = new orderByDim(thisD);
		sort(points, points+numPoints, (*myorder));
		delete myorder;
		
		
		
		list<pointPtr> leftie;
		list<pointPtr> rightie;
		int leftieSize = 0;
		int rightieSize = 0;
		int leftieMicroPoints = 0;
		int rightieMicroPoints = 0;
		for (int i=0; i<numPoints; i++){
//		for (curr=localPoints.begin(); curr!=localPoints.end(); curr++){
			rightie.push_back(points[i]);
			rightieSize++;
			if (points[i]->inMicro) rightieMicroPoints++;
		}
		
		float** leftExtent;
		float** rightExtent = findExtent(rightie);
		

		int curr = 0;
		
		float firstValue = points[0]->p[thisD];
		
		bool firstTime = true;
		
		while (curr<numPoints){
//			cerr << "#"; fflush(stderr);
//			cout << curr->first << " "; fflush(stdout);
			
//			cout << "curr :" << *(curr->second) << endl;
			
//			split = localPoints.upper_bound(curr->first);

			float splitValue = points[curr]->p[thisD];

			bool calcRightExtent = false;
			
			float nextSplitValue;
			for (int tmp=curr; tmp<numPoints; tmp++){
				if (points[tmp]->p[thisD]!=splitValue){
					nextSplitValue = points[tmp]->p[thisD];
					break;
				}
				if (tmp==numPoints-1){
					nextSplitValue = points[numPoints-1]->p[thisD];
				}
			}
			rightExtent[thisD][0] = nextSplitValue;
			
			
			while (points[curr]->p[thisD]==splitValue){
//			for (iter=curr; iter!=split; iter++){
//				cerr << "-"; fflush(stderr);
//				cout << "iter :" << *(iter->second) << endl;
				if (points[curr]->p[thisD]!=firstValue){
//				if (curr!=localPoints.begin()){
//					cout << "get updated leftExtent" << endl;
					float** temp = newExtent(leftExtent, points[curr]);
					for (int d=0; d<numDimensions; d++){
						delete[] leftExtent[d];
					}
					delete[] leftExtent;
					leftExtent = temp;
				}
				leftie.push_back(points[curr]);
				leftieSize++;
				if (points[curr]->inMicro) leftieMicroPoints++;
//				cout << "*(iter->second):" << *(iter->second) << " *(rightie.front()):" << *(rightie.front()) << endl;
				assert( (*(points[curr])==*(rightie.front())) );
				rightie.pop_front();
				rightieSize--;
				if (points[curr]->inMicro) rightieMicroPoints--;
				if (rightieSize<K){
//					cerr << "(R<K)"; fflush(stderr);
					break;
				}
				if (!calcRightExtent && !insideStr(points[curr]->p, rightExtent, numDimensions) ){
					calcRightExtent = true;
//					cerr << "*"; fflush(stderr);
				}
				curr++;		
			}
			
			if (firstTime){
				firstTime = false;
				leftExtent = findExtent(leftie);
			}
			
			if (calcRightExtent){
				for (int d=0; d<numDimensions; d++){
					delete[] rightExtent[d];
				}
				delete[] rightExtent;
				rightExtent = findExtent(rightie);
			}
			
//			cout << "leftExtent:" << getString(leftExtent, numDimensions) << " rightExtent:" << getString(rightExtent, numDimensions) << endl;
			
			
			if (leftieSize<K){
//				cerr << "(L<K)"; fflush(stderr);
				continue;
			}
			if (rightieSize<K){
//				cerr << "(R<K)"; fflush(stderr);
				break;
			}
			
			
			float score = getNCPFast(leftExtent, sizeWorld, leftieMicroPoints) + getNCPFast(rightExtent, sizeWorld, rightieMicroPoints);
			
//			float score = min( getNCPFast(leftExtent, sizeWorld, leftieMicroPoints) , getNCPFast(rightExtent, sizeWorld, rightieMicroPoints) );
			
//			float score = getNCP(leftie, sizeWorld) + getNCP(rightie, sizeWorld);
			
			
			if (score<bestScore){
				bestScore = score;
				bestDimToSplit = thisD;
				bestDimSplitValue = splitValue;
				bestLeftNumPoints = curr;
			}
			
		}
		
//		cerr << "&"; fflush(stderr);
		
		for (int d=0; d<numDimensions; d++){
			delete[] leftExtent[d];
		}
		delete[] leftExtent;
		
		for (int d=0; d<numDimensions; d++){
			delete[] rightExtent[d];
		}
		delete[] rightExtent;
		
	}
	
	
	
	
	if (bestDimToSplit<0){
//		cerr << "no split!" << endl;
		return false;
	}
	
	
	splitDimension = bestDimToSplit;
	splitValue = bestDimSplitValue;
	
	
//	cerr << "|"; fflush(stderr);

	
	// *************************************************
	// *** this code segment splits in splitValue ******

//	cout << leftNumPoints << " leftNumPoints and " << rightNumPoints << " rightNumPoints" << endl;  
	

	/*
	vector<pointPtr> indaLeft;
	vector<pointPtr> indaRight;
	
	int leftIndex=0, rightIndex=0;
	for (int i=0; i<numPoints; i++){
//		cout << "leftIndex=" << leftIndex << " rightIndex=" << rightIndex << endl;
		if (points[i]->p[splitDimension]<=splitValue){ // store them in left child
//			cout << "left" << endl;
			indaLeft.push_back( new point(*(points[i])) );
			leftIndex++;
		}
		else{ // store them in right child
//			cout << "right" << endl;
			indaRight.push_back( new point(*(points[i])) );
			rightIndex++;
		}
	}
//	cout << "leftIndex=" << leftIndex << " rightIndex" << rightIndex << endl;
	// ******************************************
	
	
	pointPtr *leftPoints, *rightPoints;
	
	
	int leftNumPoints = indaLeft.size();
	int rightNumPoints = indaRight.size();
	
	assert(leftNumPoints>=K);
	assert(rightNumPoints>=K);
	
	assert(numPoints==leftNumPoints+rightNumPoints);
	
//	cout << "numPoints=" << numPoints << " sum=" << (leftNumPoints+rightNumPoints) << " leftNumPoints=" << leftNumPoints << " rightNumPoints=" << rightNumPoints << endl; 
	
	leftPoints = new pointPtr[leftNumPoints];
	rightPoints = new pointPtr[rightNumPoints];
	
	copy(indaLeft.begin(), indaLeft.end(), leftPoints);
	copy(indaRight.begin(), indaRight.end(), rightPoints);
	*/
	
	
//	/*
	orderByDim *myorder = new orderByDim(bestDimToSplit);
	sort(points, points+numPoints, (*myorder));
	delete myorder;
	
//	cout << "bestLeftNumPoints=" << bestLeftNumPoints << endl;
	int leftNumPoints = bestLeftNumPoints;
	int rightNumPoints = numPoints-bestLeftNumPoints;
	
	pointPtr *leftPoints = new pointPtr[leftNumPoints];
	pointPtr *rightPoints = new pointPtr[rightNumPoints];
	
	int leftIndex=0, rightIndex=0;
	for (int i=0; i<numPoints; i++){
//		cerr << i << " "; fflush(stderr);
		if (i<bestLeftNumPoints){
			leftPoints[leftIndex] = points[i];
			leftIndex++;
		}
		else{
			rightPoints[rightIndex] = points[i];
			rightIndex++;
		}
	}
//	*/
	
	
	
	this->leafNode = false; // make this a non-leaf node!
	
	left = new MinPerNode(numDimensions, leftNumPoints, leftPoints, (splitDimension+1)%numDimensions);
	right = new MinPerNode(numDimensions, rightNumPoints, rightPoints, (splitDimension+1)%numDimensions);
//	cout << "ok" << endl;
	
	left->parent = this;
	right->parent = this;
	
	this->left = left;
	this->right = right;
	
	left->calculateExtentMBR();
	right->calculateExtentMBR();
	left->id = 2*id;
	right->id = 2*id + 1;
//	cout << "left child " << left;
//	cout << "right child " << right;
	
	// this code segment clears the points array for this node
//	for (int i=0; i<numPoints; i++){
//		delete[] points[i];
//	}
//	delete[] points;
	
	return true;
}



bool MinPerNode::splitNode(MinPerNodePtr &left, MinPerNodePtr &right, float* theSizeWorld, int K){
//	cerr << "splitNode (numPoints=" << numPoints << ") "; fflush(stderr);
//	if (numPoints>20*K) return fastSplitNode(left, right); // OPTIMIZATION
	
//	cout << "called splitNode for id=" << this->id << endl;
	float* sizeWorld = theSizeWorld;
	
	
	int bestDimToSplit = -1;
	float bestDimSplitValue;
	float bestScore = FLT_MAX; 
	
	for (int d=0; d<numDimensions; d++){
//		cout << "checking dimension " << d << endl;
		
		multimap<float, pointPtr> localPoints; // points ordered by float
		
		for (int i=0; i<numPoints; i++){
			localPoints.insert(pair<float, pointPtr>(points[i]->p[d], points[i]));
		}
		
//		cout << "there are " << localPoints.size() << " in the localPoints multimap" << endl;
		
		multimap<float, pointPtr>::iterator curr;
		multimap<float, pointPtr>::iterator iter;
		multimap<float, pointPtr>::iterator split;
		
		
		list<pointPtr> leftie;
		list<pointPtr> rightie;
		int leftieSize = 0;
		int rightieSize = 0;
		int leftieMicroPoints = 0;
		int rightieMicroPoints = 0;
		for (curr=localPoints.begin(); curr!=localPoints.end(); curr++){
			rightie.push_back(curr->second);
			rightieSize++;
			if (curr->second->inMicro) rightieMicroPoints++;
		}
		
		float** leftExtent;
		float** rightExtent = findExtent(rightie);
		
		
		
		
		for (curr=localPoints.begin(); curr!=localPoints.end(); curr=localPoints.upper_bound(curr->first)){
//			cerr << "#"; fflush(stderr);
//			cout << curr->first << " "; fflush(stdout);
			
//			cout << "curr :" << *(curr->second) << endl;
			
			split = localPoints.upper_bound(curr->first);
			float splitValue = curr->first;
			
			bool calcRightExtent = false;
			
			for (iter=curr; iter!=split; iter++){
//				cerr << "-"; fflush(stderr);
//				cout << "iter :" << *(iter->second) << endl;
				if (curr!=localPoints.begin()){
//					cout << "get updated leftExtent" << endl;
					float** temp = newExtent(leftExtent, iter->second);
					for (int d=0; d<numDimensions; d++){
						delete[] leftExtent[d];
					}
					delete[] leftExtent;
					leftExtent = temp;
				}
				leftie.push_back(iter->second);
				leftieSize++;
				if (iter->second->inMicro) leftieMicroPoints++;
//				cout << "*(iter->second):" << *(iter->second) << " *(rightie.front()):" << *(rightie.front()) << endl;
				assert( (*(iter->second)==*(rightie.front())) );
				rightie.pop_front();
				rightieSize--;
				if (iter->second->inMicro) rightieMicroPoints--;
				if (rightieSize<K){
//					cerr << "(R<K)"; fflush(stderr);
					break;
				}
				if ( !insideStr(iter->second->p, rightExtent, numDimensions) ){
					
					
					calcRightExtent = true;
				}			
			}
			
			if (curr==localPoints.begin()){
				leftExtent = findExtent(leftie);
			}
			
			if (calcRightExtent){
				for (int d=0; d<numDimensions; d++){
					delete[] rightExtent[d];
				}
				delete[] rightExtent;
				rightExtent = findExtent(rightie);
			}
			
//			cout << "leftExtent:" << getString(leftExtent, numDimensions) << " rightExtent:" << getString(rightExtent, numDimensions) << endl;
			
			
			if (leftieSize<K){
//				cerr << "(L<K)"; fflush(stderr);
				continue;
			}
			if (rightieSize<K){
//				cerr << "(R<K)"; fflush(stderr);
				break;
			}
			
			
			float score = getNCPFast(leftExtent, sizeWorld, leftieMicroPoints) + getNCPFast(rightExtent, sizeWorld, rightieMicroPoints);
			
//			float score = getNCP(leftie, sizeWorld) + getNCP(rightie, sizeWorld);
			
			
			if (score<bestScore){
				bestScore = score;
				bestDimToSplit = d;
				bestDimSplitValue = splitValue;
			}
			
		}
		
//		cerr << "&"; fflush(stderr);
		
		for (int d=0; d<numDimensions; d++){
			delete[] leftExtent[d];
		}
		delete[] leftExtent;
		
		for (int d=0; d<numDimensions; d++){
			delete[] rightExtent[d];
		}
		delete[] rightExtent;
		
	}
	
	
	
	
	if (bestDimToSplit<0){
//		cerr << "no split!" << endl;
		return false;
	}
	
	
	splitDimension = bestDimToSplit;
	splitValue = bestDimSplitValue;
	
	
//	cerr << "|"; fflush(stderr);

	
	/*******************************************/
	/* this code segment splits in splitValue ******/

//	cout << leftNumPoints << " leftNumPoints and " << rightNumPoints << " rightNumPoints" << endl;  
	
	
	vector<pointPtr> indaLeft;
	vector<pointPtr> indaRight;
	

	
	int leftIndex=0, rightIndex=0;
	for (int i=0; i<numPoints; i++){
//		cout << "leftIndex=" << leftIndex << " rightIndex=" << rightIndex << endl;
		if (points[i]->p[splitDimension]<=splitValue){ // store them in left child
//			cout << "left" << endl;
			indaLeft.push_back( new point(*(points[i])) );
			leftIndex++;
		}
		else{ // store them in right child
//			cout << "right" << endl;
			indaRight.push_back( new point(*(points[i])) );
			rightIndex++;
		}
	}
//	cout << "leftIndex=" << leftIndex << " rightIndex" << rightIndex << endl;
	/*******************************************/
	
	this->leafNode = false; // make this a non-leaf node!
	
	
	pointPtr *leftPoints, *rightPoints;
	
	
	int leftNumPoints = indaLeft.size();
	int rightNumPoints = indaRight.size();
	
	assert(leftNumPoints>=K);
	assert(rightNumPoints>=K);
	
	assert(numPoints==leftNumPoints+rightNumPoints);
	
//	cout << "numPoints=" << numPoints << " sum=" << (leftNumPoints+rightNumPoints) << " leftNumPoints=" << leftNumPoints << " rightNumPoints=" << rightNumPoints << endl; 
	
	leftPoints = new pointPtr[leftNumPoints];
	rightPoints = new pointPtr[rightNumPoints];
	
	copy(indaLeft.begin(), indaLeft.end(), leftPoints);
	copy(indaRight.begin(), indaRight.end(), rightPoints);
	
	
	left = new MinPerNode(numDimensions, leftNumPoints, leftPoints, (splitDimension+1)%numDimensions);
	right = new MinPerNode(numDimensions, rightNumPoints, rightPoints, (splitDimension+1)%numDimensions);
	left->parent = this;
	right->parent = this;
	
	this->left = left;
	this->right = right;
	
	left->calculateExtentMBR();
	right->calculateExtentMBR();
	left->id = 2*id;
	right->id = 2*id + 1;
//	cout << "left child " << left;
//	cout << "right child " << right;
	
	// this code segment clears the points array for this node
//	for (int i=0; i<numPoints; i++){
//		delete[] points[i];
//	}
//	delete[] points;
	
	return true;
}



bool MinPerNode::splitNodeOLD(MinPerNodePtr &left, MinPerNodePtr &right, float* theSizeWorld, int K){
	
//	if (numPoints>4*K) return fastSplitNode(left, right); // OPTIMIZATION
	
//	cout << "called splitNode for id=" << this->id << endl;
	float* sizeWorld = theSizeWorld;
	map<int, candidateSplit*> candidates; // best candidate per dimension
	
	for (int d=0; d<numDimensions; d++){
//		cout << "checking dimension " << d << endl;
		
		multimap<float, pointPtr> localPoints; // points ordered by float
		
		for (int i=0; i<numPoints; i++){
			localPoints.insert(pair<float, pointPtr>(points[i]->p[d], points[i]));
		}
		
//		cout << "there are " << localPoints.size() << " in the localPoints multimap" << endl;
		
		multimap<float, pointPtr>::iterator curr;
		multimap<float, pointPtr>::iterator iter;
		multimap<float, pointPtr>::iterator split;
		
		float thisDimBestScore = FLT_MAX;
		candidateSplit* thisDimBestCandidate;
		bool thisDimBestCandidateExists = false; // when true thisDimBestCandidate exists
		
		for (curr=localPoints.begin(); curr!=localPoints.end(); curr=localPoints.upper_bound(curr->first)){
//			cout << curr->first << " "; fflush(stdout);
			groupPtr first = new group(numDimensions);
			groupPtr second = new group(numDimensions);
			
			split = localPoints.upper_bound(curr->first);
			float splitValue = curr->first;
			
			for (iter=localPoints.begin(); iter!=split; iter++){
				first->insertPoint(*(iter->second));
			}
			for (iter=split; iter!=localPoints.end(); iter++){
				second->insertPoint(*(iter->second));
			}
			
			if (first->points.size()<(uint)K || second->points.size()<(uint)K){
				delete first;
				delete second;
			}
			else{
				candidateSplit *candSplit = new candidateSplit(splitValue, first, second, sizeWorld, K);
//				cout << "candidate " << candSplit->first->points.size() << " " << candSplit->second->points.size() << " score=" << candSplit->getScore() << endl;
				if (candSplit->getScore()<thisDimBestScore){
					if (thisDimBestCandidateExists){
						delete thisDimBestCandidate;
					}
					thisDimBestScore = candSplit->getScore();
					thisDimBestCandidate = candSplit;
					thisDimBestCandidateExists = true;
				}
			}
		}
//		cout << endl;
		
		if (thisDimBestCandidateExists){
			candidates[d] = thisDimBestCandidate; // this is the best candidate in this dimension
//			cout << "d=" << d << " " << thisDimBestCandidate->first->points.size() << " " << thisDimBestCandidate->second->points.size() << endl;
		}
	}
	
//	cout << "pick the best dimension" << endl;
	
	float bestScore = FLT_MAX;
	int bestD = -1;
	
	map<int, candidateSplit*>::iterator candIter;
	for (candIter=candidates.begin(); candIter!=candidates.end(); candIter++){
//		cout << "****** d:" << candIter->first << " score:" << candIter->second->getScore() << endl;
		if (candIter->second->getScore()<bestScore ){
			bestScore = candIter->second->getScore();
			bestD = candIter->first;
		}
	}
	if (bestD>0){
//		cout << "bestD=" << bestD << " " << candidates[bestD]->first->points.size() << " " << candidates[bestD]->second->points.size() << endl;
	}
	
	if (bestD<0){
		for (candIter=candidates.begin(); candIter!=candidates.end(); candIter++){
			delete candIter->second;
		}
//		cerr << "no split for this node" << endl;
		return false;
	}
	assert(bestD>=0);
	
	splitDimension = bestD;
	splitValue = candidates[bestD]->splitValue;
	
//	cout << "splitDimension=" << splitDimension << " splitValue=" << splitValue << endl;
	
	int leftNumPoints = candidates[bestD]->first->points.size();
	int rightNumPoints = candidates[bestD]->second->points.size();
	
	// clean up
	for (candIter=candidates.begin(); candIter!=candidates.end(); candIter++){
		delete candIter->second;
	}


// old kdtree code starts here
	
	if (rightNumPoints==0){ // don't split
//		cerr << "**** rightNumPoints==0" << endl;
		return false; // don't split
	}

	
	/*******************************************/
	/* this code segment splits in splitValue ******/

//	cout << leftNumPoints << " leftNumPoints and " << rightNumPoints << " rightNumPoints" << endl;  
	
	pointPtr *leftPoints, *rightPoints;
	
	leftPoints = new pointPtr [leftNumPoints];
	rightPoints = new pointPtr [rightNumPoints];
	
	int leftIndex=0, rightIndex=0;
	for (int i=0; i<numPoints; i++){
//		cout << "leftIndex=" << leftIndex << " rightIndex=" << rightIndex << endl;
		if (points[i]->p[splitDimension]<=splitValue){ // store them in left child
//			cout << "left" << endl;
			leftPoints[leftIndex] = new point(*(points[i]));
			leftIndex++;
		}
		else{ // store them in right child
//			cout << "right" << endl;
			rightPoints[rightIndex] = new point(*(points[i]));
			rightIndex++;
		}
	}
//	cout << "leftIndex=" << leftIndex << " rightIndex" << rightIndex << endl;
	/*******************************************/
	
	this->leafNode = false; // make this a non-leaf node!
	
	left = new MinPerNode(numDimensions, leftNumPoints, leftPoints, (splitDimension+1)%numDimensions);
	right = new MinPerNode(numDimensions, rightNumPoints, rightPoints, (splitDimension+1)%numDimensions);
	left->parent = this;
	right->parent = this;
	
	this->left = left;
	this->right = right;
	
	left->calculateExtentMBR();
	right->calculateExtentMBR();
	left->id = 2*id;
	right->id = 2*id + 1;
//	cout << "left child " << left;
//	cout << "right child " << right;
	
	// this code segment clears the points array for this node
//	for (int i=0; i<numPoints; i++){
//		delete[] points[i];
//	}
//	delete[] points;
	
	return true;
}


bool MinPerNode::fastSplitNode(MinPerNodePtr &left, MinPerNodePtr &right){
//	cout << "called fastSplitNode for id=" << this->id << endl;
	vector<float> temp;
	for (int i=0; i<numPoints; i++){
		temp.push_back(points[i]->p[splitDimension]);
	}
	
	sort(temp.begin(), temp.end()); // sort the vector!
	
	float median = *(temp.begin()+(temp.size()-1)/2);
	splitValue = median;
	
	/*******************************************/
	/* this code segment splits in median ******/
	int leftNumPoints = (temp.size()-1)/2;
	for (int i=leftNumPoints; i<numPoints; i++){
		if ( fabs(temp[i]-median)<0.001 ){
			leftNumPoints++;
		}
		else break;
	}
	
	int rightNumPoints = temp.size() - leftNumPoints;
//	cout << leftNumPoints << " leftNumPoints and " << rightNumPoints << " rightNumPoints" << endl;  
	
	pointPtr *leftPoints, *rightPoints;
	
	leftPoints = new pointPtr [leftNumPoints];
	rightPoints = new pointPtr [rightNumPoints];
	
	int leftIndex=0, rightIndex=0;
	for (int i=0; i<numPoints; i++){
//		cout << "leftIndex=" << leftIndex << " rightIndex=" << rightIndex << endl;
		if (points[i]->p[splitDimension]<=median){ // store them in left child
//			cout << "left" << endl;
			leftPoints[leftIndex] = new point(*(points[i]));
			leftIndex++;
		}
		else{ // store them in right child
//			cout << "right" << endl;
			rightPoints[rightIndex] = new point(*(points[i]));
			rightIndex++;
		}
	}
//	cout << "leftIndex=" << leftIndex << " rightIndex" << rightIndex << endl;
	/*******************************************/
	
	
	if (rightNumPoints==0){ // don't split
//		cerr << "**** rightNumPoints==0" << endl;
		return false; // don't split
	}
	
	this->leafNode = false; // make this a non-leaf node!
	
	left = new MinPerNode(numDimensions, leftNumPoints, leftPoints, (splitDimension+1)%numDimensions);
	right = new MinPerNode(numDimensions, rightNumPoints, rightPoints, (splitDimension+1)%numDimensions);
	left->parent = this;
	right->parent = this;
	
	this->left = left;
	this->right = right;
	
	left->calculateExtentMBR();
	right->calculateExtentMBR();
	left->id = 2*id;
	right->id = 2*id + 1;
//	cout << "left child " << left;
//	cout << "right child " << right;
	temp.clear();
	
	// this code segment clears the points array for this node
//	for (int i=0; i<numPoints; i++){
//		delete[] points[i];
//	}
//	delete[] points;
	
	return true;
}



float MinPerNode::getNCPFast(float** extent, float* sizeWorld, int numMicroInCluster){
//	cerr << "."; fflush(stderr);
	float perimeter = 0;
	for (int d=0; d<numDimensions; d++){
		perimeter += (extent[d][1] - extent[d][0])/sizeWorld[d];
	}
	
//	return numMicroInCluster;
	return perimeter*numMicroInCluster;
//	return perimeter;
}


float MinPerNode::getNCP(list<pointPtr> cluster, float* sizeWorld){
//	cerr << "."; fflush(stderr);
	int numDimensions = cluster.front()->numDimensions;
	float** extent = new float* [numDimensions];
	for (int d=0; d<numDimensions; d++){
		extent[d] = new float[2];
	}
	list<pointPtr>::iterator iter;
	
	int numMicroPoints = 0;
	for (iter=cluster.begin(); iter!=cluster.end(); iter++){
		if ((*iter)->inMicro){
			numMicroPoints++;
		}
		for (int d=0; d<numDimensions; d++){
			if (iter==cluster.begin()){
				extent[d][0] = (*iter)->p[d];
				extent[d][1] = (*iter)->p[d];
			}
			else{
				if ((*iter)->p[d]<extent[d][0]){
					extent[d][0] = (*iter)->p[d];
				}
				if ((*iter)->p[d]>extent[d][1]){
					extent[d][1] = (*iter)->p[d];
				}
			}
		}
	}
	
	float perimeter = 0;
	for (int d=0; d<numDimensions; d++){
		perimeter += (extent[d][1] - extent[d][0])/sizeWorld[d];
		delete[] extent[d];
	}
	delete[] extent;
	
	return perimeter;//*numMicroPoints;
}


void MinPerNode::printNode(){
	cout << this;
}

ostream& operator<<(ostream& os, const MinPerNodePtr& n){
	os << "node ID=" << n->id << " leafNode=" << n->leafNode << " ";
	os << "numDimensions=" << n->numDimensions << " numPoints=" << n->numPoints << " numMicroPoints=" << n->numMicroPoints << endl;
	os << "splitDimension=" << n->splitDimension << " splitValue=" << n->splitValue << endl;
	for (int i=0; i<n->numPoints; i++){
		os << *(n->points[i]) << " ";
	}
	os << endl;
	os << "mbr:: | ";
	for (int j=0; j<n->numDimensions; j++){
		os << n->mbr[j][0] << " -- " << n->mbr[j][1] << " | "; 
	}
	os << endl;
	os << "extent:: | ";
	for (int j=0; j<n->numDimensions; j++){
		os << n->extent[j][0] << " -- " << n->extent[j][1] << " | "; 
	}
	os << endl;
	
	return os;
}


candidateSplit::candidateSplit(float splitValue, groupPtr first, groupPtr second, float* sizeWorld, int K){
	this->K = K;
	this->splitValue = splitValue;
	this->first = first;
	this->second = second;
	numDimensions = first->numDimensions;
	this->sizeWorld = sizeWorld;
}

candidateSplit::~candidateSplit(){
	delete first;
	delete second;
}

float candidateSplit::getScore(){
	int tgt1 = 1;
	while (tgt1*K<=first->points.size()) tgt1 = tgt1<<1;
	tgt1 = tgt1>>1;
//	cout << first->points.size() << " close to " << tgt1*K << endl;
	int tgt2 = 1;
	while (tgt2*K<=second->points.size()) tgt2 = tgt2<<1;
	tgt2 = tgt2>>1;
//	cout << second->points.size() << " close to " << tgt2*K << endl;
	
	float firstScore = first->getNormalPerimeter(sizeWorld);//* (first->points.size()%K) ;//  * (1+(first->points.size()%(K*tgt1)));
	float secondScore = second->getNormalPerimeter(sizeWorld);//* (second->points.size()%K);//  * (1+(second->points.size()%(K*tgt2)));
	return (firstScore + secondScore);//*first->points.size()*second->points.size();
//	return max(firstScore, secondScore);
}


