#ifndef MinPerNODE_H_
#define MinPerNODE_H_

#include <algorithm>
#include <vector>
#include <map>
#include <list>
#include <iostream>
#include <cmath>
#include <cfloat>
#include "../common/point.h"
#include "../common/group.h"

using namespace std;


typedef class MinPerNode *MinPerNodePtr;

class MinPerNode
{
public:
int id; // node id
bool leafNode;
bool boostedLeafNode;
int numDimensions; // number of dimensions
int numPoints; // number of points
int numMicroPoints; // number of points from microdata
pointPtr *points; // the array of points
float **extent; // the extent in low-high pairs
float **mbr; // the mbr of the points in low-high pairs
int splitDimension; // the dimension along which children are split
float splitValue; // the value at which children are split

MinPerNodePtr left, right, parent; // pointers to children and parent nodes



MinPerNode(int numDimensions, int numPoints, pointPtr* points, int splitDimension); // constructor

~MinPerNode();

int calculateNumMicroPoints();

void calculateExtentMBR();
void calculateExtentMBR(float pad); // add pad around extend and mbr, for root node only
bool splitNodeNEW(MinPerNodePtr &left, MinPerNodePtr &right, float* sizeWorld, int K); // split and create children
bool splitNode(MinPerNodePtr &left, MinPerNodePtr &right, float* sizeWorld, int K); // split and create children
bool splitNodeOLD(MinPerNodePtr &left, MinPerNodePtr &right, float* sizeWorld, int K); // split and create children
bool fastSplitNode(MinPerNodePtr &left, MinPerNodePtr &right); // fast median split

void printNode();

float getNCP(list<pointPtr> cluster, float* sizeWorld);
float getNCPFast(float** extent, float* sizeWorld, int numMicroInCluster);

friend ostream& operator<<(ostream& os, const MinPerNodePtr& t);
};

ostream& operator<<(ostream& os, const MinPerNodePtr& n);


class candidateSplit
{
public:	
	int K;
	groupPtr first;
	groupPtr second;
	int numDimensions;
	float splitValue;
	float* sizeWorld;
	
	candidateSplit(float splitValue, groupPtr first, groupPtr second, float* sizeWorld, int K);
	~candidateSplit();
	
	float getScore();	
};

#endif /*MinPerNODE_H_*/
