#include "kdtree/kdtree.h"


using namespace std;

int main(int argc, char** argv){

	if (argc!=5) {
		cout << "Usage: " << argv[0] << " [microData] [publicData] [K] [selectivity]" << endl;
		cout << "K is anonymity requirement" << endl;
		exit(-1);
    }
	int K = atoi(argv[3]);
    int numDimensions;
    double selectivity = atof(argv[4]);
	vector<pointPtr> points = readFromFile(argv[1], true, numDimensions);
	Timing* tim = new Timing();

	cout << "MondrianREF" << endl;
	tim->startClock();
	KDTreePtr mdTree = new KDTree(points, K);
	long elapsed = tim->stopClock();
	cerr << "tree has " << mdTree->nodes.size() << " nodes" << endl;


//	KDTreePtr mdTree = new KDTree(argv[1], K);
//	MinPerTreePtr pub = new MinPerTree(argv[2], 2);
	KDTreePtr pub = new KDTree(argv[2], 2);
	cerr << "-------------------- trees built --------------------" << endl;

//	cout << "before boosting " << mdTree;
//	mdTree->printMetrics();
//
//	cout << endl;
//	vector<group> groups = mdTree->getGroups();
//	float* sizeWorld = mdTree->getSizeWorld();
//	calculateMetrics(groups, sizeWorld);
//	delete[] sizeWorld;
//	groups.clear();
//	cout << endl;


	cout << "boosting ..." << endl;
//	mdTree->boostWithPublicData(pub);

	tim->startClock();
	mdTree->boostWithPublicData(mdTree->root, pub);
	elapsed += tim->stopClock();
	cerr << "final tree has " << mdTree->nodes.size() << " nodes" << endl;


//	cout << "... done boosting" << endl;

//	cout << "after boosting " << mdTree;

	mdTree->calculateLeafTuples();
	mdTree->calculatePerceivedMicroTuples();


	mdTree->printMetrics();

	vector<group> groups = mdTree->getGroups();
	calculateMetrics(groups, mdTree->sizeWorld);
	groups.clear();




//	cout << "performing queries..." << endl;
	double sumError = 0, sumActual = 0, maxActual = 0, sumRelError = 0, maxRelError = 0;


	ofstream mseFile, relErrFile;
	mseFile.open("MSE.out", ios::out | ios::app);
	relErrFile.open("RelErr.out", ios::out | ios::app);

//	srand(456);


	// load points into a kdtree for getting correct answers
	KDTreePtr cor = new KDTree(argv[1], 2);

	int numQueries = 0;
	while (numQueries<100){
		double estimate=0, actual=0, estimateCOR=0, actualCOR=0;
		float **query;
		query = new float*[numDimensions];
		for (int d=0; d<numDimensions; d++){
			double rnd = ((double) rand())/RAND_MAX;
			query[d] = new float[2];
			double dimSize = mdTree->sizeWorld[d];
//			if (dimSize>1000000) dimSize = 1000000; // fix for IPUMS dataset
			query[d][0] = mdTree->nodes[1]->extent[d][0] + rnd * (1-selectivity) * dimSize;
			query[d][1] = query[d][0] + selectivity * dimSize;
//			cout << "[" << query[d][0] << " : " << query[d][1] << "] ";
		}
//		cout << endl << "query " << i << endl;
		mdTree->countQueryREF(query, estimate, actual);
		cor->countQuery(query, estimateCOR, actualCOR);
		if (actual < 1) continue; //skip this query
		cout << "... estimate=" << estimate << " actual=" << actual << endl;
		cout << "*** actualCOR=" << actualCOR << endl << endl;
		sumError += (actual-estimate)*(actual-estimate);
		sumActual += actual;
		sumRelError += fabs(actual-estimate)/actual;
		if (fabs(actual-estimate)/actual > maxRelError ){
			maxRelError = fabs(actual-estimate)/actual;
			maxActual = actual;
		}
		for (int d=0; d<numDimensions; d++){
//			cout << mdTree->sizeWorld[d] << " ";
			delete query[d];
		}
//		cout << endl;
		delete query;
		numQueries++;
	}


	mdTree->countPerceivedMicroTuples();

	cout << "AVG SQUARED ERROR " << sumError/numQueries << endl;
	mseFile << sumError/numQueries << "\t";
	cout << "AVG RELATIVE ERROR " << sumRelError/numQueries << endl;
	relErrFile << sumRelError/numQueries << "\t";
	cout << "MAX RELATIVE ERROR " << maxRelError << " at " << maxActual << endl;

	cout << "Elapsed Time: " << elapsed << endl;
	cout << "------------------------------------" << endl;
	cout << endl;

	mseFile.close();
	relErrFile.close();
}
