#ifndef STRIPES_H_
#define STRIPES_H_

#include <vector>
#include <algorithm>
#include <cmath>
#include <fstream>

#include "stripeNode.h"
#include "../common/group.h"
#include "../common/point.h"
#include "../minpertree/minpertree.h"
#include "../kdtree/kdtree.h"
#include "../common/Timing.h"


using namespace std;


class Stripes{
public:
	int numDimensions; // dimensionality
	int K; // anonymity requirement
	vector<group> groups; // K-anonymized groups
	float** extent;
	vector<pointPtr> points; // vector of pointers to all points
	float* sizeWorld;
	
	int q; // we compute q-quantiles
	
	
	stripeNodePtr root;
	
		
	
	Stripes(char *inputfile, int K);
	Stripes(vector<pointPtr> poiset, int K);
	~Stripes();
	
	
	void recursiveSplit(stripeNodePtr node);
	void recursiveSplit(stripeNodePtr node, int q, int K);
	vector<group> getGroups();
	void getGroupsRec(stripeNodePtr node);
	
	
	void run();
	float** getExtent(vector<pointPtr> poiset);
	int pickDimension(vector<pointPtr> poiset, vector<int> dims);
	void recurseStrip(vector<pointPtr> poiset, vector<int> dims);
	void createGroup(vector<pointPtr> poiset);
	void calculateSizeWorld();
	
	
	void boostWithPublicData(stripeNodePtr node, MinPerTreePtr pub);
	void boostWithPublicData(stripeNodePtr node, KDTreePtr pub);
	
	

private:
	bool groupsCalculated;
};


#endif /*STRIPES_H_*/
