#ifndef TDNODE_H_
#define TDNODE_H_

#include <algorithm>
#include <vector>
#include <map>
#include <iostream>
#include <cmath>
#include <cfloat>
#include "../common/point.h"
#include "../common/group.h"

using namespace std;


typedef class TDNode *TDNodePtr;

class TDNode
{
public:
int id; // node id
bool isLeftChild;
bool leafNode;
bool boostedLeafNode;
int numDimensions; // number of dimensions
int numPoints; // number of points
int numMicroPoints; // number of points from microdata
pointPtr *points; // the array of points
float **extent; // the extent in low-high pairs
float **mbr; // the mbr of the points in low-high pairs
int splitDimension; // the dimension along which children are split
float splitValue; // the value at which children are split
double leafTuples; // the number of all tuples under this node
double perceivedMicroTuples; // the perceived number volume of all MT tuples at this leaf; only for leaves

TDNodePtr left, right, parent; // pointers to children and parent nodes



TDNode(int numDimensions, int numPoints, pointPtr* points, int splitDimension); // constructor

~TDNode();

int calculateNumMicroPoints();

void calculateExtentMBR();
void calculateExtentMBR(float pad); // add pad around extend and mbr, for root node only
bool splitNode(TDNodePtr &left, TDNodePtr &right, float* sizeWorld, int K); // split and create children
bool fastSplitNode(TDNodePtr &left, TDNodePtr &right); // fast median split

float getNCP(vector<pointPtr> cluster, pointPtr po, float* sizeWorld);
float getNCP(vector<pointPtr> cluster, float* sizeWorld);

float getNCPIncrease(float** oldExtent, float** &nExtent, float oldNCP, pointPtr po, int oldNumMicroPoints,  float* sizeWorld);

void printNode();

friend ostream& operator<<(ostream& os, const TDNodePtr& t);
};

ostream& operator<<(ostream& os, const TDNodePtr& n);



class bucket{
public:
	vector<pointPtr> points;
	float** mbr;
	TDNodePtr node;
	bool altered;
	int numMicroPoints;

	bucket(){
		altered = false;
	}

	~bucket(){
		int numDims = points[0]->numDimensions;
		for (int i=0; i<numDims; i++){
			delete[] mbr;
		}
		delete[] mbr;
	}

	float getNCP(float* sizeWorld){
		return numMicroPoints*getNormalizedPerimeter(mbr, sizeWorld, points[0]->numDimensions);
	}

};


class candidateSplit
{
public:
	int K;
	groupPtr first;
	groupPtr second;
	int numDimensions;
	float splitValue;
	float* sizeWorld;

	candidateSplit(float splitValue, groupPtr first, groupPtr second, float* sizeWorld, int K);
	~candidateSplit();

	float getScore();
};

#endif /*TDNODE_H_*/
