#ifndef TDTree_H_
#define TDTree_H_

#include <fstream>
#include <sstream>
#include <map>
#include <set>
#include <string>
#include <algorithm>
#include <cmath>
#include "tdnode.h"
#include "../common/group.h"
#include "../kdtree/kdtree.h"
#include "../common/Timing.h"

using namespace std;


typedef class TDTree *TDTreePtr;

class TDTree
{
public:
	int numDimensions; // dimensionality

	map<int, TDNodePtr> nodes;
//	vector<NodePtr> nodes;

	int K; // anonymity requirement

	vector<group> groups; // K-anonymized groups

	ofstream fout;
	TDNodePtr root;

	float* sizeWorld;

	TDTree(char *inputfile, int K);
	TDTree(vector<pointPtr> poiset, int K);

	void recursiveSplit(TDNodePtr node);

	void calculateSizeWorld();

	void writeGroups(char *inputfile);
	vector<group> getGroups();
	void printMetrics();

	void filter();

	vector<pointPtr> rangeQuery(float **query); // construct VO
	void recRangeQuery(TDNodePtr node, float **query, vector<pointPtr>& res); // recursive descend

	// strict versions:
	vector<pointPtr> rangeQueryStr(float **query); // construct VO
	void recRangeQueryStr(TDNodePtr node, float **query, vector<pointPtr>& res); // recursive descend

	void boostWithPublicData(TDNodePtr node, KDTreePtr pub);
	void boostWithPublicData(TDTreePtr pub);


	void calculateLeafTuples(); // calculates the aggregate leafTuples at each node
	void recCalculateLeafTuples(TDNodePtr node);

	void calculatePerceivedMicroTuples(); // calculates the perceived MT tuples at the leaves
	void recCalculatePerceivedMicroTuples(TDNodePtr node, double ratio);


	void countQuery(float **query, double& estimate, double& actual); // performs count range query
	void recCountQuery(TDNodePtr node, float **query, double& estimate, double& actual);

	void countQueryREF(float **query, double& estimate, double& actual); // performs count range query for REFINEMENT
	void recCountQueryREF(TDNodePtr node, float **query, double& estimate, double& actual);


	void countPerceivedMicroTuples();
	void recCountPerceivedMicroTuples(TDNodePtr node, double& count, double& actual);


	void markLeaves();
	void recMarkLeaves(TDNodePtr node);

	void calculateMicroTuples();
	void recCalculateMicroTuples(TDNodePtr node, double& numMicroTuples);


	friend ostream& operator<<(ostream& os, const TDTree*& t);
	friend ostream& operator<<(ostream& os, const TDTree& t);
//	friend bool overlaps(float **rect1, float **rect2);
//	friend bool inside(float* point, float ** rect);

private:
	bool groupsCalculated;
};

ostream& operator<<(ostream& os, const TDTreePtr& t);
ostream& operator<<(ostream& os, const TDTree& t);

string getString(float* point, int numDimensions);
bool overlaps(float **rect1, float **rect2, int numDimensions);
bool inside(float* point, float ** rect, int numDimensions);
// strict version:
bool overlapsStr(float **rect1, float **rect2, int numDimensions);
bool insideStr(float* point, float ** rect, int numDimensions);


#endif /*TDTree_H_*/
