#include "tdtree/tdtree.h"


using namespace std;

int main(int argc, char** argv){

	if (argc!=5) {
		cout << "Usage: " << argv[0] << " [microData] [publicData] [K] [selectivity]" << endl;
		cout << "K is anonymity requirement" << endl;
		exit(-1);
    }
	int K = atoi(argv[3]);
	int numDimensions;
	double selectivity = atof(argv[4]);
	vector<pointPtr> points = readFromFile(argv[1], true, numDimensions);
	Timing* tim = new Timing();


    cout << "TopDownREF" << endl;
//	TDTreePtr mdTree = new TDTree(argv[1], K);
	tim->startClock();
	TDTreePtr mdTree = new TDTree(points, K);
	long splitTime = tim->stopClock();

	tim->startClock();
	mdTree->filter();
	long initialFilterTime = tim->stopClock();

	mdTree->markLeaves();

	mdTree->calculateMicroTuples();


	KDTreePtr pub = new KDTree(argv[2], 2);
	cerr << "-------------------- trees built --------------------" << endl;

    tim->startClock();
    mdTree->boostWithPublicData(mdTree->root, pub);
    long boostTime = tim->stopClock();

    tim->startClock();
    mdTree->filter(); // IMPORTANT
    long finalFilterTime = tim->stopClock();

    mdTree->markLeaves();

    vector<group> groups = mdTree->getGroups();
	calculateMetrics(groups, mdTree->sizeWorld);
	groups.clear();



	mdTree->calculateLeafTuples();
	mdTree->calculatePerceivedMicroTuples();


//	cout << "performing queries..." << endl;
	double sumError = 0, sumActual = 0, maxActual = 0, sumRelError = 0, maxRelError = 0;


	ofstream mseFile, relErrFile;
	mseFile.open("MSE.out", ios::out | ios::app);
	relErrFile.open("RelErr.out", ios::out | ios::app);




// load points into a kdtree for getting correct answers
	KDTreePtr cor = new KDTree(argv[1], 2);

//	srand(456);

	int numQueries = 0;
	while (numQueries<100){
		double estimate=0, actual=0, estimateCOR=0, actualCOR=0;
		float **query;
		query = new float*[numDimensions];
		for (int d=0; d<numDimensions; d++){
			double rnd = ((double) rand())/RAND_MAX;
			query[d] = new float[2];
			double dimSize = mdTree->sizeWorld[d];
//			if (dimSize>1000000) dimSize = 1000000; // fix for IPUMS dataset
			query[d][0] = mdTree->nodes[1]->extent[d][0] + rnd * (1-selectivity) * dimSize;
			query[d][1] = query[d][0] + selectivity * dimSize;
//			cout << "[" << query[d][0] << " : " << query[d][1] << "] ";
		}
//		cout << endl << "query " << i << endl;
		mdTree->countQueryREF(query, estimate, actual);
		cor->countQuery(query, estimateCOR, actualCOR);
		if (actualCOR < 1) continue; //skip this query
		cout << "... estimate=" << estimate << " actual=" << actual << endl;
		cout << "*** actualCOR=" << actualCOR << endl << endl;
		sumError += (actualCOR-estimate)*(actualCOR-estimate);
		sumActual += actualCOR;
		sumRelError += fabs(actualCOR-estimate)/actualCOR;
		if (fabs(actualCOR-estimate)/actualCOR > maxRelError ){
			maxRelError = fabs(actualCOR-estimate)/actualCOR;
			maxActual = actualCOR;
		}
		for (int d=0; d<numDimensions; d++){
//			cout << mdTree->sizeWorld[d] << " ";
			delete query[d];
		}
//		cout << endl;
		delete query;
		numQueries++;
	}


	mdTree->countPerceivedMicroTuples();


	cout << "AVG SQUARED ERROR " << sumError/numQueries << endl;
	mseFile << sumError/numQueries << "\t";
	cout << "AVG RELATIVE ERROR " << sumRelError/numQueries << endl;
	relErrFile << sumRelError/numQueries << "\t";
	cout << "MAX RELATIVE ERROR " << maxRelError << " at " << maxActual << endl;

	cout << "Split Time: " << splitTime << " initial Filter Time: " << initialFilterTime;
	cout << " Boost Time: " << boostTime << " final Filter Time: " << finalFilterTime << endl;
	cout << "------------------------------------" << endl;
	cout << endl;


	mseFile.close();
	relErrFile.close();

/*
	cout << "TopDownREF" << endl;
	TDTreePtr mdTree = new TDTree(argv[1], K);
	TDTreePtr pub = new TDTree(argv[2], 1);

//	cout << "before boosting " << mdTree;
//	mdTree->printMetrics();
//
//	cout << endl;
//	vector<group> groups = mdTree->getGroups();
//	calculateMetrics(groups, mdTree->sizeWorld);
//	groups.clear();
//	cout << endl;


//	cout << "boosting ..." << endl;
	mdTree->boostWithPublicData(pub);
//	cout << "... done boosting" << endl;

//	cout << "after boosting " << mdTree;
	mdTree->printMetrics();

	vector<group> groups = mdTree->getGroups();
	calculateMetrics(groups, mdTree->sizeWorld);
	groups.clear();
	cout << "------------------------------------" << endl;
	cout << endl;
*/
}
